#include <algorithm>
#include <memory>
#include <src/server/file/FileServer.h>
#include <log/LogUtils.h>
#include <misc/std_unique_ptr.h>
#include <pipes/buffer.h>
#include "FileClient.h"

using namespace std;
using namespace std::chrono;
using namespace ts::server;

void FileClient::sendRawMessage(const pipes::buffer_view &message) {
    lock_guard lock(this->bufferLock);
    this->write_queue.push_back(message.own_buffer());

    if(!this->event_write_hold)
        event_add(this->writeEvent, nullptr);
    else {
        __asm__("nop");
    }
}

void FileClient::handleMessageWrite(int fd, short, void *) {
    auto self = this->_this.lock();
    if(!self) return;

    decltype(this->pendingKey) pending_key;
    {

        lock_guard<threads::Mutex> l(this->tickLock);
        pending_key = this->pendingKey;
    }

    unique_lock lock(this->bufferLock, defer_lock);
    if(!lock.try_lock_for(milliseconds(5))) {
        logWarning(LOG_FT, "{} Buffer lock locked, could not write data!", this->client_prefix());
        return; /* somebody else doing a action and will (hopefully) readd the event */
    }
    if(self->state_connection == C_DISCONNECTED) return;

    auto used = this->used_bandwidth();
    if(pending_key) { //Delay the write <3 :P
        if(pending_key->max_bandwhidth >= 0 && used > pending_key->max_bandwhidth) {
            event_write_hold = true;
            logTrace(LOG_FT, "{} Exceeded bandwidth limit of {} bytes (Used {} bytes). Temporary skipping write event!", this->client_prefix(), pending_key->max_bandwhidth, this->used_bandwidth());
            return;
        }
    }
    event_write_hold = false;

    pipes::buffer* buffer = nullptr;
    while(true) {
        if(this->write_queue.empty()) {
            lock.unlock(); /* unlock write buffer because we're ticking */
            if(pending_key && !pending_key->upload)
                this->handle->tickFileClient(_this.lock()); //We have to fill up again
            return;
        }

        buffer = &this->write_queue.front();
        if(buffer->empty()) {
            this->write_queue.pop_front();
            buffer = nullptr;
            continue;
        }
        break;
    }

    ssize_t length = buffer->length();
    if(pending_key && pending_key->max_bandwhidth >= 0) {
        if(pending_key->max_bandwhidth < length && pending_key->max_bandwhidth > used) length = pending_key->max_bandwhidth - used;
    }

    length = send(fd, buffer->data_ptr(), length, MSG_NOSIGNAL);
    //logTrace(LOG_FT, "{} Wrote {} bytes", this->client_prefix(), length);

    if(length == -1) {
        if (errno == EINTR || errno == EAGAIN) {
            event_add(this->writeEvent, nullptr);
            return;
        }
        else {
            logError(LOG_FT, "{} Failed to write some data! ({}/{})", this->client_prefix(), errno, strerror(errno));
            lock.unlock();
            self->disconnect();
            return;
        }
    } else {
        *buffer = buffer->range(length);

        auto entry = make_unique<BandwidthEntry>();
        entry->length = length ;
        entry->timestamp = system_clock::now();
        {
            lock_guard<recursive_mutex> b_lock(this->bandwidth_lock);
            this->bandwidth.push_back(move(entry));
        }
    }
    last_io_action = system_clock::now();

    if(buffer->empty()) {
        this->write_queue.pop_front();
        buffer = nullptr;
    }

    if(pending_key && this->bytesHandled != pending_key->size) {
        if(this->write_queue.size() < 4) this->handle->tickFileClient(_this.lock());
    }

    if(!this->write_queue.empty()) {
        event_add(this->writeEvent, nullptr);
    }
    self.reset();
}

void FileClient::handleMessageRead(int fd, short, void *) {
    auto self = this->_this.lock();
    if(self->state_connection != C_CONNECTED) return;

    decltype(this->pendingKey) pending_key;
    {

        lock_guard<threads::Mutex> l(this->tickLock);
        pending_key = this->pendingKey;
    }
    size_t buffer_length = 1024;
    if(pending_key && pending_key->max_bandwhidth >= 0) {
        auto used = this->used_bandwidth();
        if(used < pending_key->max_bandwhidth) {
            buffer_length = pending_key->max_bandwhidth - used;
            if(buffer_length > 1024) buffer_length = 1024;
        } else {
            logTrace(LOG_FT, "{} Exceeded bandwidth limit of {} bytes (Used {} bytes). Temporary removing read event!", this->client_prefix(), pending_key->max_bandwhidth, used);
            {
                lock_guard lock(this->bufferLock);
                if(this->readEvent)
                    event_del_noblock(this->readEvent);
            }
            this->event_read_hold = true;
            return;
        }
    }
    pipes::buffer buffer(buffer_length);
    auto length = recv(fd, buffer.data_ptr(), buffer.length(), 0);
    if(length < 0){
        if(errno == EINTR || errno == EAGAIN)
            ;//event_add(this->readEvent, nullptr);
        else {
            {
                lock_guard lock(this->bufferLock);
                if(this->readEvent)
                    event_del_noblock(this->readEvent);
            }
            if(this->state_connection == C_CONNECTED) {
                logError(LOG_FT, "{} Failed to read some data! ({}/{})", this->client_prefix(), errno, strerror(errno));
                self->disconnect();
            }
        }
        return;
    } else if(length == 0){
        {
            lock_guard lock(this->bufferLock);
            if(this->readEvent)
                event_del_noblock(this->readEvent);
        }
        if(this->state_connection == C_CONNECTED) {
            if(this->state_transfer == T_TRANSFER)
                logWarning(LOG_FT, "{} Transfer hang up. Remote peer closed the connection.", this->client_prefix());
            else
                logMessage(LOG_FT, "{} Remote peer has closed the connection before initializing a transfer.", this->client_prefix());
            self->disconnect(seconds(3));
        }
        return;
    }
    last_io_action = system_clock::now();

    buffer.resize(length);
    {
        lock_guard lock(this->bufferLock);
        if(self->state_connection != C_CONNECTED) return; /* drop the buffer because we're not connected anymore */

        this->read_queue.push_back(std::move(buffer));
    }

    this->handle->tickFileClient(_this.lock());

    auto entry = make_unique<BandwidthEntry>();
    entry->length = length ;
    entry->timestamp = system_clock::now();
    //logTrace(LOG_FT, "{} Readed {} bytes", this->client_prefix(), length);
    {
        lock_guard<recursive_mutex> lock(this->bandwidth_lock);
        this->bandwidth.push_back(move(entry));
    }
}


size_t FileClient::availableBytes() {
    lock_guard lock(this->bufferLock);
    size_t available = 0;
    for(const auto& buf : this->read_queue)
        available += buf.length();
    return available;
}

std::string FileClient::peekBytes(size_t size) {
    ssize_t required = size;
    lock_guard lock(this->bufferLock);

    string result;
    result.reserve(size);

    for(pipes::buffer& buf : this->read_queue) {
        if(required <= 0) break;
        if(buf.length() > required) {
            result += string(buf.data_ptr<const char>(), required);
            required = 0;
        } else {
            result += string(buf.data_ptr<const char>(), buf.length());
            required -= buf.length();
        }
    }
    return result;
}

std::string FileClient::getBytes(size_t size) {
    lock_guard lock(this->bufferLock);

    string result;
    result.reserve(size);

    while(!this->read_queue.empty()) {
        if(size <= 0) break;

        auto& buf = this->read_queue.front();
        if(buf.length() > size) {
            result += string((char*) buf.pipes::buffer_view::data_ptr(), size);
            buf = buf.range(size);
            size = 0;
        } else {
            result += string((char*) buf.data_ptr(), buf.length());
            size -= buf.length();
            this->read_queue.pop_front();
        }
    }
    return result;
}