#pragma once

#include <vector>
#include <string>
#include <chrono>
#include <memory>
#include <ThreadPool/Thread.h>
#include <deque>
#include <ThreadPool/Mutex.h>
#include <netinet/in.h>
#include <sqlite3.h>
#include <sql/SqlQuery.h>
#include "../Group.h"
#include <event.h>
#include <misc/net.h>
#include "../manager/IpListManager.h"

namespace ts {
    class BasicChannel;
    class Group;

    namespace server {
        class VirtualServer;
        class QueryClient;

        struct QueryLoginCredentials {
            std::string username;
            std::string password;
            ClientUid uniqueId;
        };

        class QueryAccount {
            public:
                std::string username;
                std::string unique_id;

                ServerId bound_server;

                QueryAccount() = default;
                virtual ~QueryAccount() = default;
        };

        class PasswortedQueryAccount : public QueryAccount {
            public:
                std::string password;

                PasswortedQueryAccount() = default;
                virtual ~PasswortedQueryAccount() = default;
        };

        class QueryServer {
                friend class QueryClient;
            public:
                struct Binding {
                    sockaddr_storage address{};
                    int file_descriptor = 0;
                    ::event* event_accept = nullptr;

                    inline std::string as_string() { return net::to_string(address, true); }
                };

                explicit QueryServer(sql::SqlManager*);
                ~QueryServer();

                bool start(const std::deque<std::shared_ptr<Binding>>& /* bindings */, std::string& /* error */);
                void stop();
                bool running(){ return active; }

                void unregisterConnection(const std::shared_ptr<QueryClient> &);

                std::deque<std::shared_ptr<QueryAccount>> list_query_accounts(OptionalServerId /* server */);
                std::shared_ptr<QueryAccount> create_query_account(const std::string& /* name */, ServerId /* server */, const std::string& /* owner unique id */, const std::string& /* password */);
                std::shared_ptr<PasswortedQueryAccount> load_password(const std::shared_ptr<QueryAccount>& /* account */);
                bool delete_query_account(const std::shared_ptr<QueryAccount>& /* account */);

                std::shared_ptr<QueryAccount> find_query_account_by_name(const std::string& /* name */);
                std::deque<std::shared_ptr<QueryAccount>> find_query_accounts_by_unique_id(const std::string &unique_id /* unique id */);

                bool rename_query_account(const std::shared_ptr<QueryAccount>& /* account */, const std::string& /* new name */);
                bool change_query_password(const std::shared_ptr<QueryAccount>& /* account */, const std::string& /* new password */);

                threads::ThreadPool* executePool() { return this->_executePool; }
            private:
                sql::SqlManager* sql;

                bool active = false;
                std::deque<std::shared_ptr<Binding>> bindings;
                std::vector<threads::Thread*> threads;

                std::mutex server_reserve_fd_lock;
                int server_reserve_fd = -1; /* -1 = unset | 0 = in use | > 0 ready to use */

                std::unique_ptr<IpListManager> ip_whitelist;
                std::unique_ptr<IpListManager> ip_blacklist;

                //IO stuff
                event_base* eventLoop = nullptr;
                std::chrono::system_clock::time_point accept_event_deleted;

                threads::ThreadPool* _executePool = nullptr;

                std::mutex connected_clients_lock;
                std::deque<std::shared_ptr<QueryClient>> connectedClients;

                threads::Mutex loginLock;
                std::chrono::system_clock::time_point lastDecrease;
                std::map<std::string, uint32_t> loginAttempts;
                std::map<std::string, std::chrono::system_clock::time_point> queryBann;

                threads::Thread* ioThread = nullptr;
                threads::SchedulingTask tickingId = nullptr;
                void on_client_receive(int fd, short ev, void *arg);
                void tick();
        };
    }
}