#include "FileServer.h"
#include "src/client/file/FileClient.h"
#include "src/client/ConnectedClient.h"
#include <netinet/tcp.h>
#include <experimental/filesystem>
#include <log/LogUtils.h>
#include <misc/std_unique_ptr.h>
#include "src/InstanceHandler.h"

using namespace std;
using namespace std::chrono;
using namespace ts;
using namespace ts::file;
using namespace ts::server;
namespace fs = std::experimental::filesystem;

#if defined(TCP_CORK) && !defined(TCP_NOPUSH)
    #define TCP_NOPUSH TCP_CORK
#endif

extern InstanceHandler* serverInstance;

FileServer::FileServer() {}

FileServer::~FileServer() {
    stop();
}

inline fs::path buildPath(std::string rootPath, std::shared_ptr<file::FileEntry> directory){
    auto strPath = directory ? directory->path + "/" + directory->name : "";
    if(strPath.find(rootPath) == 0) return fs::u8path(strPath);
    return fs::u8path(rootPath + strPath);
}

std::shared_ptr<file::Directory> FileServer::createDirectory(std::string name, std::shared_ptr<file::Directory> parent) {
    auto path = buildPath(this->rootPath, parent);
    path += name;
    if(!fs::exists(path))
        if(!fs::create_directories(path)) return nullptr;
        else ;
    else if(!fs::is_directory(path)) return nullptr;

    return static_pointer_cast<file::Directory>(this->findFile(path.string()));
}

bool FileServer::fileExists(std::shared_ptr<file::Directory> dir) {
    return fs::exists(buildPath(this->rootPath, dir));
}

bool FileServer::fileExists(std::shared_ptr<file::File> file) {
    return fs::exists(buildPath(this->rootPath, file));
}

std::shared_ptr<file::FileEntry> FileServer::findFile(std::string path, std::shared_ptr<file::Directory> parent) {
    if(path.find(this->rootPath) != 0) {
        string strPath = (parent ? parent->path + "/" + parent->name : this->rootPath) + "/";
        if(strPath.find(this->rootPath) != 0 && rootPath != strPath)
            path = this->rootPath + strPath;
        if(path.find(strPath) == 0 && !strPath.empty())
            ;
        else path = strPath + path;
    }

    fs::path absPath = fs::u8path(path);
    if(!fs::is_regular_file(absPath) && !fs::is_directory(absPath)){
        debugMessage(LOG_FT, "Could not find requested file. Abs path: {} | {}. (path={}, parent={})", absPath.string(), path, path, (parent ? parent->path + "/" + parent->name : "./"));
        return nullptr;
    }

    std::shared_ptr<file::FileEntry> entry;
    if(fs::is_directory(absPath))
        entry = std::make_shared<file::Directory>();
    else entry = std::make_shared<file::File>();

    entry->name = absPath.filename();
    entry->type = fs::is_directory(absPath) ? FileType::DIRECTORY : FileType::FILE;
    entry->path = absPath.parent_path().string();
    entry->lastChanged = fs::last_write_time(absPath);
    if(entry->type == FileType::FILE)
        static_pointer_cast<file::File>(entry)->fileSize = entry->type == FileType::FILE ? fs::file_size(absPath) : 0;
    return entry;
}

std::vector<std::shared_ptr<file::FileEntry>> FileServer::listFiles(std::shared_ptr<file::Directory> dir) {
    if(!dir) return {};
    auto directory = buildPath(this->rootPath, dir);

    if(!fs::exists(directory)) return {};
    if(!fs::is_directory(directory)) return {};

    std::vector<std::shared_ptr<file::FileEntry>> result;
    deque<fs::directory_entry> files;
    for(const auto& elm : fs::directory_iterator(directory)) {
        files.push_back(elm);
    }
    std::sort(files.begin(), files.end(), [](const fs::directory_entry& a, const fs::directory_entry& b) {
        return fs::last_write_time(a.path()) > fs::last_write_time(b.path());
    });
    for(const auto& elm : files) {
        if(fs::is_regular_file(elm.path())){
            auto entry = make_shared<file::File>();
            entry->name = elm.path().filename();
            entry->type = FileType::FILE;
            entry->path = elm.path().parent_path().string();
            entry->lastChanged = fs::last_write_time(elm.path());
            entry->fileSize = fs::file_size(elm.path());
            result.push_back(entry);
        } else if(fs::is_directory(elm.path())){
            auto entry = make_shared<file::Directory>();
            entry->name = elm.path().filename();
            entry->type = FileType::DIRECTORY;
            entry->path = elm.path().parent_path().string();
            entry->lastChanged = fs::last_write_time(elm.path());
            result.push_back(entry);
        } else {
            logError(LOG_FT, "Invalid file in file tree. File path: " + elm.path().string());
        }
    }

    return result;
}

bool FileServer::deleteFile(std::shared_ptr<file::FileEntry> file) {
    return fs::remove_all(fs::u8path(file->path + "/" + file->name)) > 0;
}

inline std::string randomString(uint length = 15, std::string charIndex = "abcdefghijklmnaoqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890")
{
    std::string rs = "";
    for (uint i = 0; i < length; ++i)
        rs += charIndex[rand() % charIndex.length()];
    return rs;
}

std::shared_ptr<file::FileTransfereKey> FileServer::generateDownloadTransferKey(std::string &errorMessage, std::string targetFile, size_t offset, const shared_ptr<ConnectedClient>& client) {
    auto file = this->findFile(targetFile, nullptr);
    if(!file){
        errorMessage = "file does not exists";
        return nullptr;
    }
    if(file->type != FileType::FILE) {
        errorMessage = "file is a directory";
        return nullptr;
    }

    auto result = make_shared<file::FileTransfereKey>();
    result->owner = client;
    result->server = client->getServerId();

    result->key_id = (uint16_t) (rand() & 0xFFFF);
    result->createTimestamp = std::chrono::system_clock::now();
    result->key = randomString(16);
    result->offset = offset;
    result->size = static_pointer_cast<file::File>(file)->fileSize;
    result->upload = false;
    result->targetFile = targetFile;

    { //Test for server?
        auto bandwidth_server = client->getServer()->properties()[property::VIRTUALSERVER_MAX_UPLOAD_TOTAL_BANDWIDTH].as<int64_t>();
        auto bandwidth_instance = serverInstance->properties()[property::SERVERINSTANCE_MAX_UPLOAD_TOTAL_BANDWIDTH].as<int64_t>();

        result->max_server_bandwidth = bandwidth_server;
        result->max_bandwhidth = bandwidth_server;

        if(bandwidth_instance != -1) {
            if(bandwidth_instance < bandwidth_server)
                result->max_bandwhidth = bandwidth_instance; //We want to limit that already until exact calculations was made
        }
    }

    if(client->getServer())
        result->max_server_bandwidth = client->getServer()->properties()[property::VIRTUALSERVER_MAX_DOWNLOAD_TOTAL_BANDWIDTH];

    {
        threads::MutexLock lock(this->keylock);
        pendingKeys.push_back(result);
    }
    return result;
}

std::shared_ptr<file::FileTransfereKey> FileServer::generateUploadTransferKey(std::string &errorMessage, std::string targetFile, size_t size, size_t offset, const shared_ptr<ConnectedClient>& client) {
    shared_ptr<file::FileTransfereKey> result = make_shared<file::FileTransfereKey>();
    result->owner = client;
    result->server = client->getServerId();

    result->key_id = (uint16_t) (rand() & 0xFFFF);
    result->createTimestamp = std::chrono::system_clock::now();
    result->key = randomString(16);
    result->offset = offset;
    result->size = size;
    result->upload = true;
    result->targetFile = targetFile;

    { //Test for server?
        auto bandwidth_server = client->getServer()->properties()[property::VIRTUALSERVER_MAX_DOWNLOAD_TOTAL_BANDWIDTH].as<int64_t>();
        auto bandwidth_instance = serverInstance->properties()[property::SERVERINSTANCE_MAX_DOWNLOAD_TOTAL_BANDWIDTH].as<int64_t>();

        result->max_server_bandwidth = bandwidth_server;
        result->max_bandwhidth = bandwidth_server;

        if(bandwidth_instance != -1) {
            if(bandwidth_instance < bandwidth_server)
                result->max_bandwhidth = bandwidth_instance; //We want to limit that already until exact calculations was made
        }
    }
    {
        threads::MutexLock lock(this->keylock);
        pendingKeys.push_back(result);
    }
    debugMessage(LOG_FT, "Created file upload key=" + result->key + " for " + targetFile + " (" + to_string(size) + " bytes)");
    return result;
}

std::shared_ptr<file::Directory> FileServer::resolveDirectory(const shared_ptr<VirtualServer> &server, std::shared_ptr<BasicChannel> channel, std::string subPath) {
    fs::path path = fs::u8path("server_" + to_string(server ? server->getServerId() : 0) + "/channel_" + to_string(channel->channelId()));
    if(!findFile(path))
        this->createDirectory(path.string(), nullptr);
    path += subPath;

    auto ffile = findFile(path.string());
    debugMessage(LOG_FT, "Resolve {} => {} -> {}", path.string(), (void*) ffile.get(), typeid(ffile).name());
    return static_pointer_cast<file::Directory>(ffile);
}

std::shared_ptr<file::Directory> FileServer::iconDirectory(const shared_ptr<VirtualServer> &server) {
    fs::path root = fs::u8path(this->rootPath);
    fs::path path = fs::u8path("server_" + to_string(server ? server->getServerId() : 0) + "/icons");
    if(!fs::exists(root / path)) {
        if(!fs::create_directories(root / path))
            return nullptr;
    }
    return static_pointer_cast<file::Directory>(findFile(path.string()));
}

bool FileServer::iconExists(const shared_ptr<VirtualServer> &server, IconId icon) {
    if(icon == 0) return false;
    if(icon < 1000) return true;
    return this->findFile("icon_" + to_string(icon), this->iconDirectory(server)) != nullptr;
}

std::shared_ptr<file::Directory> FileServer::avatarDirectory(const shared_ptr<VirtualServer> &server) {
    fs::path path = fs::u8path("server_" + to_string(server ? server->getServerId() : 0) + "/avatars");
    if(!findFile(path))
        this->createDirectory(path.string(), nullptr);
    return static_pointer_cast<file::Directory>(findFile(path.string()));
}

#define TS3_ICON_HASH "icon_1001"
void FileServer::setupServer(const shared_ptr<VirtualServer> &server) {
    auto dir = iconDirectory(server);
    if(!dir) {
        logError(LOG_FT,"Failed to find icon directory for server {}", server ? server->getServerId() : 0);
    } else {
        if(!fs::exists(fs::u8path(dir->path + "/" + dir->name + "/" TS3_ICON_HASH))) {
            try {
                fs::copy(fs::u8path("resources/teaspeak16px.png"), fs::u8path(dir->path + "/" + dir->name + "/" + TS3_ICON_HASH));
            } catch(std::exception& ex) {
                logError(LOG_FT, "Failed to copy default path: {}", ex.what());
            }
        }
    }
}

std::string FileServer::server_file_base(const std::shared_ptr<ts::server::VirtualServer> &server) {
    return rootPath + "/server_" + to_string(server->getServerId());
}

void FileServer::deleteServer(const shared_ptr<VirtualServer> &server) {
    fs::path path = fs::u8path(rootPath + "/server_" + to_string(server ? server->getServerId() : 0));
    if(fs::exists(path)) {
        error_code error;
        if(fs::remove_all(path, error) == 0)
            logError(LOG_FT, "Could not delete server directory {} ({} | {})", path.string(), error.value(), error.message());
    } else {
        logError(LOG_FT, "Could not delete missing server directory (" + path.string() + ")");
    }
}

//The actual server!
bool FileServer::start(const std::deque<std::shared_ptr<FileServer::Binding>>& bindings, std::string& error) {
    if(this->running()) {
        error = "server already running";
        return false;
    }
    this->active = true;

    /* reserve backup file descriptor in case that the max file descriptors have been reached  */
    {
        this->server_reserve_fd = dup(1);
        if(this->server_reserve_fd < 0)
            logWarning(LOG_FT, "Failed to reserve a backup accept file descriptor. ({} | {})", errno, strerror(errno));
    }

    /* setup event bases */
    {
        this->ioLoop = event_base_new();
        this->ioThread = new threads::Thread(THREAD_SAVE_OPERATIONS | THREAD_EXECUTE_LATER, [&]{
            while(this->active) {
                debugMessage(LOG_FT, "Entering event loop ({})", (void*) this->ioLoop);
                event_base_loop(this->ioLoop, EVLOOP_NO_EXIT_ON_EMPTY);
                if(this->active) {
                    debugMessage(LOG_FT, "Event loop exited ({}). No active events. Sleeping 1 seconds", (void*) this->ioLoop);
                    this_thread::sleep_for(seconds(1));
                } else {
                    debugMessage(LOG_FT, "Event loop exited ({})", (void*) this->ioLoop);
                }
            }
        });
        this->ioThread->name("File IO #1").execute();
    }

    {
        for(auto& binding : bindings) {
            binding->file_descriptor = socket(binding->address.ss_family, SOCK_STREAM | SOCK_NONBLOCK, 0);
            if(binding->file_descriptor < 0) {
                logError(LOG_FT, "Failed to bind server to {}. (Failed to create socket: {} | {})", binding->as_string(), errno, strerror(errno));
                continue;
            }

            int enable = 1, disabled = 0;

            if (setsockopt(binding->file_descriptor, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(int)) < 0)
                logWarning(LOG_FT, "Failed to activate SO_REUSEADDR for binding {} ({} | {})", binding->as_string(), errno, strerror(errno));
            if(setsockopt(binding->file_descriptor, IPPROTO_TCP, TCP_NOPUSH, &disabled, sizeof disabled) < 0)
                logWarning(LOG_FT, "Failed to deactivate TCP_NOPUSH for binding {} ({} | {})", binding->as_string(), errno, strerror(errno));
            if(binding->address.ss_family == AF_INET6) {
                if(setsockopt(binding->file_descriptor, IPPROTO_IPV6, IPV6_V6ONLY, &enable, sizeof(int)) < 0)
                    logWarning(LOG_FT, "Failed to activate IPV6_V6ONLY for IPv6 binding {} ({} | {})", binding->as_string(), errno, strerror(errno));
            }
            if(fcntl(binding->file_descriptor, F_SETFD, FD_CLOEXEC) < 0)
                logWarning(LOG_FT, "Failed to set flag FD_CLOEXEC for binding {} ({} | {})", binding->as_string(), errno, strerror(errno));


            if (bind(binding->file_descriptor, (struct sockaddr *) &binding->address, sizeof(binding->address)) < 0) {
                logError(LOG_FT, "Failed to bind server to {}. (Failed to bind socket: {} | {})", binding->as_string(), errno, strerror(errno));
                close(binding->file_descriptor);
                continue;
            }

            if (listen(binding->file_descriptor, SOMAXCONN) < 0) {
                logError(LOG_FT, "Failed to bind server to {}. (Failed to listen: {} | {})", binding->as_string(), errno, strerror(errno));
                close(binding->file_descriptor);
                continue;
            }

            binding->event_accept = event_new(this->ioLoop, binding->file_descriptor, EV_READ | EV_PERSIST, [](int a, short b, void* c){ ((FileServer *) c)->on_client_accept(a, b, c); }, this);
            event_add(binding->event_accept, nullptr);
            this->bindings.push_back(binding);
        }

        if(this->bindings.empty()) {
            this->stop();
            error = "failed to bind to any address";
            return false;
        }
    }

    for(int index = 0; index < 2; index++){
        auto th = new threads::Thread(THREAD_SAVE_OPERATIONS | THREAD_EXECUTE_LATER, &FileServer::clientTickingExecutor, this);
        th->name("Ticking FT #" + to_string(index)).execute();
        this->tickingThreads.push_back(th);
    }
    return true;
}

void FileServer::stop() {
    if(!this->running()) return;
    active = false;
    {
        lock_guard<mutex> lock(this->tickingLock);
        this->tickingCon.notify_all();
    }

    for(auto& binding : this->bindings) {
        if(binding->event_accept) {
            event_del_block(binding->event_accept);
            event_free(binding->event_accept);
            binding->event_accept = nullptr;
        }
        if(binding->file_descriptor > 0) {
            if(shutdown(binding->file_descriptor, SHUT_RDWR) < 0)
                logWarning(LOG_FT, "Failed to shutdown socket for binding {} ({} | {}).", binding->as_string(), errno, strerror(errno));
            if(close(binding->file_descriptor) < 0)
                logError(LOG_FT, "Failed to close socket for binding {} ({} | {}).", binding->as_string(), errno, strerror(errno));
            binding->file_descriptor = -1;
        }
    }
    this->bindings.clear();

    auto clClone = this->connectedClients;
    for(const auto& cl : clClone) {
        cl->disconnect(chrono::milliseconds(1));
    }

    if(this->ioLoop) {
        event_base_loopbreak(this->ioLoop);
        event_base_loopexit(this->ioLoop, nullptr);
    }

    for(const auto& thread : this->tickingThreads){
        if(thread->join(chrono::seconds(1)) != 0) logCritical(LOG_FT, "Failed to terminal file server tick thread!");
        delete thread;
    }
    this->tickingThreads.clear();

    if(this->ioThread)
        if(this->ioThread->join(chrono::system_clock::now() + chrono::seconds(3)) != 0) {
            logCritical(LOG_FT, "Could not terminate the event base dispatch thread of the file server");
            this->ioThread->detach();
        }
    delete this->ioThread;
    this->ioThread = nullptr;

    if(this->ioLoop) {
        event_base_free(this->ioLoop);
        this->ioLoop = nullptr;
    }

    if(this->server_reserve_fd > 0) {
        if(close(this->server_reserve_fd) < 0)
            logError(LOG_FT, "Failed to close backup file descriptor ({} | {})", errno, strerror(errno));
    }
    this->server_reserve_fd = -1;
}

inline std::string logging_address(const sockaddr_storage& address) {
    if(config::server::disable_ip_saving)
        return "[0|X.X.X.X" + to_string(net::port(address)) + "|unconnected]";
    return "[0|X.X.X.X" + net::to_string(address, true) + "|unconnected]";
}

#define CLOSE_CONNECTION \
if(shutdown(file_descriptor, SHUT_RDWR) < 0) { \
    debugMessage(LOG_FT, "[{}] Failed to shutdown socket ({} | {}).", logging_address(remote_address), errno, strerror(errno)); \
} \
if(close(file_descriptor) < 0) { \
    debugMessage(LOG_FT, "[{}] Failed to close socket ({} | {}).", logging_address(remote_address), errno, strerror(errno)); \
}

void FileServer::on_client_accept(int _server_file_descriptor, short ev, void *arg) {
    sockaddr_storage remote_address{};
    memset(&remote_address, 0, sizeof(remote_address));
    socklen_t address_length = sizeof(remote_address);

    int file_descriptor = accept(_server_file_descriptor, (struct sockaddr *) &remote_address, &address_length);
    if (file_descriptor < 0) {
        if(errno == EAGAIN)
            return;

        if(errno == EMFILE || errno == ENFILE) {
            if(errno == EMFILE)
                logError(LOG_FT, "Server ran out file descriptors. Please increase the process file descriptor limit or decrease the instance variable 'serverinstance_filetransfer_max_connections'");
            else
                logError(LOG_FT, "Server ran out file descriptors. Please increase the process and system-wide file descriptor limit or decrease the instance variable 'serverinstance_filetransfer_max_connections'");

            bool tmp_close_success = false;
            {
                lock_guard reserve_fd_lock(server_reserve_fd_lock);
                if(this->server_reserve_fd > 0) {
                    debugMessage(LOG_FT, "Trying to accept client with the reserved file descriptor to close the incomming connection.");
                    auto _ = [&]{
                        if(close(this->server_reserve_fd) < 0) {
                            debugMessage(LOG_FT, "Failed to close reserved file descriptor");
                            tmp_close_success = false;
                            return;
                        }
                        this->server_reserve_fd = 0;

                        errno = 0;
                        file_descriptor = accept(_server_file_descriptor, (struct sockaddr *) &remote_address, &address_length);
                        if(file_descriptor < 0) {
                            if(errno == EMFILE || errno == ENFILE)
                                debugMessage(LOG_FT, "[{}] Even with freeing the reserved descriptor accept failed. Attempting to reclaim reserved file descriptor", logging_address(remote_address));
                            else if(errno == EAGAIN);
                            else {
                                debugMessage(LOG_FT, "[{}] Failed to accept client with reserved file descriptor. ({} | {})", logging_address(remote_address), errno, strerror(errno));
                            }
                            this->server_reserve_fd = dup(1);
                            if(this->server_reserve_fd < 0)
                                debugMessage(LOG_FT, "[{}] Failed to reclaim reserved file descriptor. Future clients cant be accepted!", logging_address(remote_address));
                            else
                                tmp_close_success = true;
                            return;
                        }
                        debugMessage(LOG_FT, "[{}] Successfully accepted client via reserved descriptor (fd: {}). Disconnecting client.", logging_address(remote_address), file_descriptor);

                        CLOSE_CONNECTION
                        this->server_reserve_fd = dup(1);
                        if(this->server_reserve_fd < 0)
                            debugMessage(LOG_FT, "Failed to reclaim reserved file descriptor. Future clients cant be accepted!");
                        else
                            tmp_close_success = true;
                        logMessage(LOG_FT, "[{}] Dropping file transfer connection attempt because of too many open file descriptors.", logging_address(remote_address));
                    };
                    _();
                }
            }

            if(!tmp_close_success) {
                debugMessage(LOG_FT, "Sleeping two seconds because we're currently having no resources for this user. (Removing the accept event)");
                for(auto& binding : this->bindings)
                    event_del_noblock(binding->event_accept);
                accept_event_deleted = system_clock::now();
                return;
            }
            return;
        }
        logMessage(LOG_FT, "Got an error while accepting a new client. (errno: {}, message: {})", errno, strerror(errno));
        return;
    }
    {
        unique_lock lock(this->clientLock);
        auto max_connections = serverInstance->properties()[property::SERVERINSTANCE_FILETRANSFER_MAX_CONNECTIONS].as<size_t>();
        if(max_connections > 0 && max_connections <= this->connectedClients.size()) {
            lock.unlock();
            logMessage(LOG_FT, "[{}] Dropping new connection attempt because of too many connected clients.", logging_address(remote_address));
            CLOSE_CONNECTION
            return;
        }

        auto max_ip_connections = serverInstance->properties()[property::SERVERINSTANCE_FILETRANSFER_MAX_CONNECTIONS_PER_IP].as<size_t>();
        if(max_ip_connections > 0) {
            size_t connection_count = 0;
            for(auto& client : this->connectedClients) {
                if(net::address_equal(client->remote_address, remote_address))
                    connection_count++;
            }

            if(connection_count >= max_ip_connections) {
                lock.unlock();
                logMessage(LOG_FT, "[{}] Dropping new connection attempt because of too many simultaneously connected session from this ip.", logging_address(remote_address));
                CLOSE_CONNECTION
                return;
            }
        }
    }


    shared_ptr<FileClient> client = std::make_shared<FileClient>(this, file_descriptor);
    client->_this = client;
    memcpy(&client->remote_address, &remote_address, sizeof(remote_address));

    this->clientLock.lock();
    this->connectedClients.push_back(client);
    this->clientLock.unlock();
    event_add(client->readEvent, nullptr);
    logMessage(LOG_FT, "[{}] Remote peer connected. Initializing session.", logging_address(remote_address));
}

void FileServer::clientTickingExecutor() {
    while(this->running()){
        shared_ptr<FileClient> client;
        {
            unique_lock<mutex> lock(this->tickingLock);
            this->tickingCon.wait(lock, [&](){ return !this->running() || !this->tickQueue.empty(); });
            if(!this->running()) return;

            client = this->tickQueue.front();
            this->tickQueue.pop_front();
        }

        if(client->tick()) this->tickFileClient(client); //Client needs more ticking
    }
}

std::deque<std::shared_ptr<FileClient>> FileServer::running_file_transfers(const std::shared_ptr<ts::server::ConnectedClient> &client) {
    std::deque<std::shared_ptr<FileClient>> result;

    {
        threads::MutexLock lock(this->clientLock);
        for(const auto& c : this->connectedClients) {
            if(!client || c->client == client) result.push_back(c);
        }
    }

    return result;
}

std::deque<std::shared_ptr<file::FileTransfereKey>> FileServer::pending_file_transfers(const std::shared_ptr<ts::server::ConnectedClient> &client) {
    std::deque<std::shared_ptr<file::FileTransfereKey>> result;
    for(const auto& key : this->pending_keys())
        if(!client || key->owner.lock() == client)
            result.push_back(key);
    return result;
}

void FileServer::tickFileClient(std::shared_ptr<FileClient> cl) {
    lock_guard<mutex> lock(this->tickingLock);
    this->tickQueue.push_back(cl);
    this->tickingCon.notify_one();
}

struct TransfareGroup {
    shared_ptr<VirtualServer> server;
    deque<shared_ptr<FileClient>> clients;
    enum {
        upload,
        download
    } direction = upload;
    size_t used_bandwidth = 0;
    ssize_t max_bandwidth = -1;
};

void FileServer::instanceTick() {
    {
        //tickQueue
        auto client = this->connected_clients();
        lock_guard<mutex> tick_lock(this->tickingLock);
        this->tickQueue.insert(this->tickQueue.end(), client.begin(), client.end()); //Tick all clients :)
        this->tickingCon.notify_all();
    }
    if(this->accept_event_deleted.time_since_epoch().count() != 0 && accept_event_deleted + seconds(5) < system_clock::now()) {
        debugMessage(LOG_FT, "Readding accept event and try again if we have enough resources again.");
        for(auto& binding : this->bindings)
            event_add(binding->event_accept, nullptr);
        accept_event_deleted = system_clock::time_point{};
    }

    auto now = system_clock::now();
    if(timestamp_bandwidth_update + seconds(1) < now) {
        timestamp_bandwidth_update = now;

        for(const auto& entry : this->pending_keys()) {
            if(!entry) continue;
            if(entry->createTimestamp + minutes(1) < now) {
                logMessage(entry->server, "[FILE] Timeout file transfer for file " + entry->targetFile + "! Dropping key");

                {
                    threads::MutexLock lock(this->keylock);
                    auto index = find(this->pendingKeys.begin(), this->pendingKeys.end(), entry);
                    if(index != this->pendingKeys.end())
                        this->pendingKeys.erase(index);
                }
            }
        }

        deque<shared_ptr<TransfareGroup>> groups;
        {
            for(const auto& client : this->connected_clients()) {
                if(!client || !client->client) continue;

                lock_guard<threads::Mutex> l(client->tickLock); //Instance hangup here
                if(!client->pendingKey) continue;

                auto used_band = client->used_bandwidth();
                if(client->pendingKey && client->state_connection != FileClient::C_DISCONNECTED && (client->pendingKey->max_bandwhidth < 0 || used_band < client->pendingKey->max_bandwhidth)) {

                    if(client->event_read_hold) { //Reactivate read!
                        event_add(client->readEvent, nullptr);
                        client->event_read_hold = false;
                        debugMessage(LOG_FT, "{} Reattaching client read event! (Current bandwidth {})", client->client_prefix(), used_band);
                    }
                    if(client->event_write_hold) { //Reactivate read!
                        event_add(client->writeEvent, nullptr);
                        client->event_write_hold = false;
                        debugMessage(LOG_FT, "{} Reattaching client write event! (Current bandwidth {})", client->client_prefix(), used_band);
                    }
                }
                if(!client->client->getServer()) continue;

                bool found = false;
                for(auto& group : groups) {
                    if (group->server == client->client->getServer() && (group->direction == TransfareGroup::upload) == client->pendingKey->upload) {
                        group->clients.push_back(client);
                        group->used_bandwidth += client->used_bandwidth();
                        found = true;
                        break;
                    }
                }
                if(found) continue;

                auto entry = make_shared<TransfareGroup>();
                entry->direction = client->pendingKey->upload ? TransfareGroup::upload : TransfareGroup::download;
                entry->used_bandwidth = client->used_bandwidth();
                entry->clients.push_back(client);
                entry->server = client->client->getServer();
                groups.push_back(entry);
            }
        }

        { //Instance limit upload
            auto limit_upload = serverInstance->properties()[property::SERVERINSTANCE_MAX_UPLOAD_TOTAL_BANDWIDTH].as<ssize_t>();
            if(limit_upload >= 0) {
                deque<shared_ptr<TransfareGroup>> elements;
                size_t used_bandwidth = 0;
                for(const auto& entry : groups) {
                    if (entry->direction == TransfareGroup::upload) {
                        elements.push_back(entry);
                        used_bandwidth += entry->used_bandwidth;
                    }
                }
                if(!elements.empty() && used_bandwidth > limit_upload) {
                    for(auto& entry : elements) {
                        entry->max_bandwidth = ssize_t(ceil(double(entry->used_bandwidth) * double(limit_upload) / double(used_bandwidth))); //Adjust the total
                    }
                }
            }
        }

        { //Instance limit download
            auto limit_download = serverInstance->properties()[property::SERVERINSTANCE_MAX_DOWNLOAD_TOTAL_BANDWIDTH].as<ssize_t>();
            if(limit_download >= 0) {
                deque<shared_ptr<TransfareGroup>> elements;
                size_t used_bandwidth = 0;
                for(const auto& entry : groups) {
                    if (entry->direction == TransfareGroup::download) {
                        elements.push_back(entry);
                        used_bandwidth += entry->used_bandwidth;
                    }
                }
                if(!elements.empty() && used_bandwidth > limit_download) {
                    for(auto& entry : elements) {
                        entry->max_bandwidth = ssize_t(ceil(double(entry->used_bandwidth) * double(limit_download) / double(used_bandwidth))); //Adjust the total
                    }
                }
            }
        }

        map<ServerId, deque<shared_ptr<TransfareGroup>>> servers;
        for(const auto& entry : groups) {
            servers[entry->server->getServerId()].push_back(entry);
        }

        for(const auto& server : servers) {
            if(server.second.empty()) continue;

            { //Adjust server upload
                auto max_bandwidth = server.second.front()->server->properties()[property::VIRTUALSERVER_MAX_UPLOAD_TOTAL_BANDWIDTH].as<ssize_t>();
                if(max_bandwidth >= 0) {
                    deque<shared_ptr<TransfareGroup>> tranfares;
                    size_t bandwidth = 0;
                    for(const auto& trans : server.second) {
                        if(trans->direction == TransfareGroup::upload) {
                            tranfares.push_back(trans);
                            bandwidth += trans->max_bandwidth;
                        }
                    }

                    if(bandwidth > max_bandwidth) {
                        for(auto& trans : tranfares)
                            trans->max_bandwidth = ssize_t(ceil(double(trans->max_bandwidth) * double(max_bandwidth) / double(bandwidth))); //Adjust the total
                    }
                }
            }

            { //Adjust server download
                auto max_bandwidth = server.second.front()->server->properties()[property::VIRTUALSERVER_MAX_DOWNLOAD_TOTAL_BANDWIDTH].as<ssize_t>();
                if(max_bandwidth >= 0) {
                    deque<shared_ptr<TransfareGroup>> tranfares;
                    size_t bandwidth = 0;
                    for(const auto& trans : server.second)
                        if(trans->direction == TransfareGroup::download) {
                            tranfares.push_back(trans);
                            bandwidth += trans->max_bandwidth;
                        }

                    if(bandwidth > max_bandwidth) {
                        for(auto& trans : tranfares)
                            trans->max_bandwidth = ssize_t(ceil(double(trans->max_bandwidth) * double(max_bandwidth) / double(bandwidth))); //Adjust the total
                    }
                }
            }
        }
    }
}