#include "src/VirtualServer.h"
#include "TeamSpeakWebClient.h"
#include "log/LogUtils.h"
#include "src/server/VoiceServer.h"

using namespace ts;
using namespace ts::weblist;
using namespace ts::server;

TSWebClient::TSWebClient(const std::shared_ptr<VirtualServer> &server, struct event_base *event_base, uint16_t session_index, bool resend_name) : server(server), event_base(event_base), session_index(session_index), send_name(resend_name) { }

TSWebClient::~TSWebClient() {
    this->unregister_events(true);
    try {
        if(this->close_thread.get_id() == std::this_thread::get_id())
            this->close_thread.detach();
        else if(this->close_thread.joinable())
            this->close_thread.join();
    } catch(std::exception& ex) {
        logCritical(this->server ? this->server->getServerId() : 0, "[WebList] Failed to join or detach close thread! message: {}", ex.what());
    }
}

void TSWebClient::_handle_message_read(int file_descriptor, short, void *ptr_client) {
    const auto client = (TSWebClient*) ptr_client;

    pipes::buffer buffer{512};
    const auto read = recv(file_descriptor, buffer.data_ptr(), buffer.length(), MSG_NOSIGNAL | MSG_DONTWAIT);
    if(read <= 0) {
        logError(client->server->getServerId(), "[WebList] Failed to read weblist response. ({} | {} => {})", read, errno, strerror(errno));
        event_del_noblock(client->event_read);
        client->trigger_fail_later("failed to read!", true);
        return;
    }

    client->handle_message_read(buffer.view(0, read));
}

void TSWebClient::_handle_message_write(int file_descriptor, short, void *ptr_client) {
    const auto client = (TSWebClient*) ptr_client;
    pipes::buffer buffer;

    {
        std::lock_guard lock(client->write_lock);
        if(client->write_buffer.empty()) return;
        buffer = std::move(client->write_buffer[0]);
        client->write_buffer.pop_front();

        if(!client->write_buffer.empty() && client->event_write)
            event_add(client->event_write, nullptr);
    }

    const auto write = sendto(file_descriptor, buffer.data_ptr(), buffer.length(), 0, (const sockaddr *) &client->remote_address, sizeof(sockaddr_in));
    if(write != buffer.length()){
        logError(client->server->getServerId(), "[WebList] Failed to send weblist response. ({} | {} => {})", write, errno, strerror(errno));
        client->trigger_fail_later("failed to write!", true);
        return;
    }
}

void TSWebClient::_handle_timeout(int, short, void *ptr_client) {
    const auto client = (TSWebClient*) ptr_client;
    if(client->event_timeout)
        event_del_noblock(client->event_timeout);
    client->trigger_fail_later("timeout", true);
}

void TSWebClient::unregister_events(bool blocking) {
    const auto unregister = [blocking](event*& event) {
        if(blocking) {
            event_del_block(event);
            event_free(event);
            event = nullptr;
        } else {
            event_del_noblock(event);
        }
    };

    if(this->event_timeout)
        unregister(this->event_timeout);
    if(this->event_read)
        unregister(this->event_read);
    if(this->event_write)
        unregister(this->event_write);

    if(this->file_descriptor > 0) {
        close(this->file_descriptor);
        this->file_descriptor = 0;
    }
}

void TSWebClient::reset_timeout(bool reschedule) {
    if(this->event_timeout) {
        event_del(this->event_timeout);
        if(reschedule) {
            timeval timeout{0};
            timeout.tv_sec = 5;
            event_add(this->event_timeout, &timeout);
        }
    }
}

void TSWebClient::trigger_fail_later(const std::string &error, bool retry) {
    if(this->state >= S_FINALIZING || this->close_thread.joinable()) return;
    this->state = S_FINALIZING;
    this->close_thread = std::thread([&, error, retry](){
        this->unregister_events(true);
        if(this->callback_error)
            this->callback_error(error,retry);
    });
}

void TSWebClient::trigger_success_later() {
    if(this->state >= S_FINALIZING || this->close_thread.joinable()) return;
    this->state = S_FINALIZING;
    this->close_thread = std::thread([&](){
        this->unregister_events(true);
        if(this->callback_success)
            this->callback_success();
    });
}

void TSWebClient::write_message(const pipes::buffer_view &buffer) {
    if(this->state >= S_FINALIZING) return;

    {
        std::lock_guard lock(this->write_lock);
        this->write_buffer.push_back(buffer.own_buffer());
    }

    if(this->event_write)
        event_add(this->event_write, nullptr);
}

#define BASE_PORT 30566
void TSWebClient::report() {
    this->state = S_INITIALIZING;

    memset(&this->remote_address, 0, sizeof(sockaddr_in));
    this->remote_address.sin_family = AF_INET;
    this->remote_address.sin_port = htons(2010);
    {
        auto record = gethostbyname("weblist.teamspeak.com");
        if(!record) {
            this->trigger_fail_later("Failed to resolve weblist.teamspeak.com", true);
            return;
        }
        this->remote_address.sin_addr.s_addr = ((in_addr*) record->h_addr)->s_addr;
    }


    this->file_descriptor = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(this->file_descriptor <= 0) {
        this->trigger_fail_later("Failed to create file descriptor", true);
        return;
    }

    if(fcntl(this->file_descriptor, F_SETFD, FD_CLOEXEC) < 0) {
        this->trigger_fail_later("Failed to enable FD_CLOEXEC for {} (WebReport)", true);
        return;
    }

    int allow = 1;
    setsockopt(this->file_descriptor, SOL_SOCKET, SO_REUSEADDR, &allow, sizeof(int));

    {
        sockaddr_in address{};
        address.sin_family = AF_INET;
        address.sin_port = htons(BASE_PORT + this->server->getServerId());
        address.sin_addr.s_addr = INADDR_ANY;

        if(this->server->getVoiceServer()) {
            for(const auto& binding : this->server->getVoiceServer()->activeBindings()) {
                if(binding->address.ss_family != AF_INET) continue;

                address.sin_addr.s_addr = ((sockaddr_in*) &binding->address)->sin_addr.s_addr;
                debugMessage(this->server->getServerId(), "[WebList] Requesting from local address {}:{}", binding->address_string(), ntohs(address.sin_port));
                break;
            }
        }

        if(bind(this->file_descriptor, (const sockaddr *) &address, sizeof(address)) < 0) {
            this->trigger_fail_later("Failed to bind socket", true);
            return;
        }
    }

    this->event_read = event_new(this->event_base, this->file_descriptor, EV_READ | EV_PERSIST, TSWebClient::_handle_message_read, this);
    this->event_write = event_new(this->event_base, this->file_descriptor, EV_WRITE, TSWebClient::_handle_message_write, this);
    this->event_timeout = event_new(this->event_base, -1, 0, TSWebClient::_handle_timeout, this);
    event_add(this->event_read, nullptr);

    this->reset_timeout(true);
    this->request_session();
}

void TSWebClient::abort() {
    this->trigger_fail_later("aborted", false);
}

void TSWebClient::abort_sync() {
    this->unregister_events(true);
    if(this->callback_error)
        this->callback_error("aborted", false);
}

inline char nibble(int c) {
    assert(c < 16);

    if(c >= 0 && c <= 9)
        return (char) ('0' + c);
    else
        return (char) ('A' + (c - 10));
}

inline std::string hex(const pipes::buffer_view &buffer) {
    std::string result;
    for(size_t index = 0; index < buffer.length(); index++)
        result += std::string() + "0x" + nibble(buffer[index] >> 4) + nibble(buffer[index] & 0x0F) + " ";

    if(result.empty()) return result;
    return result.substr(0, result.length() - 1);
}

void TSWebClient::handle_message_read(const pipes::buffer_view &buffer) {
    this->reset_timeout(true);
    if(this->state == S_INITIALIZING) {
        if(buffer.length() != 8) {
            this->trigger_fail_later("Failed to receive key. (Invalid packet length of " + std::to_string(buffer.length()) + ". Expected 8. Buffer: " + hex(buffer) + ")", true);
            return;
        }

        this->session_id = *(uint32_t*) &buffer[4];
        this->send_status(this->send_name);
    } else if(this->state == S_DATA_EXCHANGE) {
        if(buffer.length() != 5) {
            this->trigger_fail_later("Failed to receive exchange stats. (Invalid packet length of " + std::to_string(buffer.length()) + ". Expected 5. Buffer: " + hex(buffer) + ")", true);
            return;
        }

        int status = buffer[4];
        if(status == 7) {
            this->send_name = true;
            debugMessage(this->server->getServerId(), "[WebList] Received exchange status 7. Resending with server name.");
            this->send_status(true);
            return;
        } else if(status == 0) {
            this->trigger_success_later();
            return;
        } else {
            debugMessage(this->server->getServerId(), "[WebList] Weblist request failed. Status code {}. Buffer: {}", status, hex(buffer));
            this->trigger_fail_later("Invalid exchange status. Status: " + std::to_string(status), status != 5);
            return;
        }
    }
}

void TSWebClient::request_session() {
    this->state = S_INITIALIZING;

    pipes::buffer buffer{4};
    buffer[0] = 1; //Version
    buffer[1] = (uint8_t) (session_index >> 8); //Packet number
    buffer[2] = (uint8_t) (session_index & 0xFF);
    buffer[3] = 1; //Request session

    this->write_message(buffer);
}

void TSWebClient::send_status(bool flag_name) {
    this->state = S_DATA_EXCHANGE;

    auto name = this->server->getDisplayName();
    pipes::buffer buffer{4 + 4 + 3 * 2 + 1 + 1 + (flag_name ? name.length() : 0)};

    buffer[0] = 1; //Version
    buffer[1] = (uint8_t) (session_index >> 8); //Packet number
    buffer[2] = (uint8_t) (session_index & 0xFF);
    buffer[3] = 2; //Status update

    *(uint32_t*) &buffer[4] = this->session_id;
    *(uint16_t*) &buffer[8] = this->server->properties()[property::VIRTUALSERVER_PORT].as<uint16_t>();
    *(uint16_t*) &buffer[10] = this->server->properties()[property::VIRTUALSERVER_MAXCLIENTS].as<uint16_t>();
    *(uint16_t*) &buffer[12] = this->server->properties()[property::VIRTUALSERVER_CLIENTS_ONLINE].as<uint16_t>();

    bool flag_channel = this->server->could_default_create_channel();
    bool flag_password = this->server->properties()[property::VIRTUALSERVER_FLAG_PASSWORD].as<bool>();
    debugMessage(this->server->getServerId(), "[WebList] Sending properties: flag_channel: {}, flag_password: {}", flag_channel, flag_password);
    buffer[14] = (flag_password ? 0x01 : 0) | (flag_channel ? 0x02 : 0); //Create channels (0x01 = password)

    if(flag_name) {
        buffer[15] = (uint8_t) name.length();
        if(name.length() > 0) {
            memcpy(&buffer[16], name.c_str(), name.length());
        }
    } else {
        buffer[15] = 0; //We dont want to send a server name
    }

    this->write_message(buffer);
}